﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Optimization.Operators.LCS {
  [Item("ILASOperator", "Description missing")]
  [StorableClass]
  public class ILASOperator : SingleSuccessorOperator, IStochasticOperator {

    #region Parameter Properties
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<IGAssistProblemData> ProblemDataParameter {
      get { return (ILookupParameter<IGAssistProblemData>)Parameters["ProblemData"]; }
    }
    public IValueLookupParameter<IntValue> NumberOfStrataParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["NumberOfStrata"]; }
    }
    public IValueLookupParameter<ItemList<ItemList<IntValue>>> StrataParameter {
      get { return (IValueLookupParameter<ItemList<ItemList<IntValue>>>)Parameters["Strata"]; }
    }
    #endregion

    [StorableConstructor]
    protected ILASOperator(bool deserializing) : base(deserializing) { }
    protected ILASOperator(ILASOperator original, Cloner cloner)
      : base(original, cloner) {
    }
    public ILASOperator()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random generator to use."));
      Parameters.Add(new LookupParameter<IGAssistProblemData>("ProblemData", ""));
      Parameters.Add(new ValueLookupParameter<IntValue>("NumberOfStrata", ""));
      Parameters.Add(new ValueLookupParameter<ItemList<ItemList<IntValue>>>("Strata", ""));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ILASOperator(this, cloner);
    }

    public override IOperation Apply() {
      if (StrataParameter.ActualValue == null) {
        //initialize strata
        int numberOfStrata = NumberOfStrataParameter.ActualValue.Value;
        var strata = new ItemList<ItemList<IntValue>>(numberOfStrata);
        for (int i = 0; i < numberOfStrata; i++) {
          strata.Add(new ItemList<IntValue>());
        }

        //get all rows for every niche
        var problemData = ProblemDataParameter.ActualValue;
        IntRange evaluate = ProblemDataParameter.ActualValue.TrainingPartition;
        Dictionary<IGAssistNiche, List<int>> nicheRows = null;
        for (int row = evaluate.Start; row < evaluate.End; row++) {
          if (problemData.IsTrainingSample(row)) {
            var niche = problemData.FetchAction(row);
            if (nicheRows == null) {
              nicheRows = new Dictionary<IGAssistNiche, List<int>>(niche.Comparer);
            }
            if (!nicheRows.ContainsKey(niche)) {
              nicheRows.Add(niche, new List<int>());
            }
            nicheRows[niche].Add(row);
          }
        }

        //distribute niches to strata
        var random = RandomParameter.ActualValue;
        foreach (var niche in nicheRows.Keys) {
          var rows = nicheRows[niche];
          int count = 0;
          while (rows.Count > 0) {
            int pos = random.Next(rows.Count);
            strata[count % numberOfStrata].Add(new IntValue(rows[pos]));
            rows.RemoveAt(pos);
            count++;
          }
        }

        if (strata.Any(x => x.Count() == 0)) {
          throw new ArgumentException("At least one strata is empty. Use less strata or add more data.");
        }

        StrataParameter.ActualValue = strata;
      }
      return base.Apply();
    }
  }
}
